@@ -1,7 +1,10 @@
 
 
+import chunk
 import typing
 import collections
+import math
+import codecs
 #import datetime
 
 from ._Hex import _Hex
@@ -20,6 +23,8 @@ class DumperSettings(object):
 		self.showComplexStructsWithType = False
 		self.compactSequenceLengthLimit = 8
 		self.compactSequenceItemLengthLimit = 50
+		self.bytesLineSize = 32
+		self.compactBytesLinesLengthLimit = 32
 	#
 
 #
@@ -76,6 +81,114 @@ def _int_toBits(data:int) -> typing.Union[int,str]:
 	return _Bits(data)
 #
 
+def _byteChunker(data:bytes, chunkSize:int) -> typing.Sequence[bytes]:
+	assert isinstance(data, bytes)
+	assert isinstance(chunkSize, int)
+	assert chunkSize > 0
+
+	iFrom = 0
+	iTo = chunkSize
+	while iFrom < len(data):
+		yield data[iFrom:iTo]
+		iFrom = iTo
+		iTo += chunkSize
+#
+
+def _x_byteChunkToHex(data:bytes) -> str:
+	hexSpanLength = 16		# == 8 bytes for a span
+
+	s = codecs.encode(data, "hex").decode("ascii")
+	chunks = [ s[i:i+hexSpanLength] for i in range(0, len(s), hexSpanLength) ]
+	return " ".join(chunks)
+#
+
+def _x_byteChunkToASCII(data:bytes) -> str:
+	spanLength = 8		# == 8 bytes for a span
+
+	ret = []
+	for i, b in enumerate(data):
+		if (i % spanLength) == 0:
+			ret.append(" ")
+		if 32 <= b <= 127:
+			ret.append(chr(b))
+		else:
+			ret.append(".")
+
+	return "".join(ret)
+#
+
+#
+# Returns chunks of the specified data.
+#
+def _byteChunkerWithOfs(settings:DumperSettings, data:bytes, processorName:str = None) -> typing.Sequence[typing.Tuple[str,str,str]]:
+	assert isinstance(settings, DumperSettings)
+	assert isinstance(data, bytes)
+	chunkSize = settings.bytesLineSize
+	assert isinstance(chunkSize, int)
+	assert chunkSize > 0
+	hexStrPadded = chunkSize*2 + math.ceil(chunkSize / 8) - 1
+	if processorName is not None:
+		assert isinstance(processorName, str)
+
+	# ----
+
+	nTotalLength = len(data)
+	nTotalLines = math.ceil(nTotalLength / chunkSize)
+	formatStrFragment = None
+	formatStrFragmentEllipsis = None
+	if nTotalLength <= 256*256:
+		formatStrFragment = "{:04x}"
+		formatStrFragmentEllipsis = "... "
+	elif nTotalLength <= 256*256*256:
+		formatStrFragment = "{:06x}"
+		formatStrFragmentEllipsis = "...   "
+	else:
+		formatStrFragment = "{:08x}"
+		formatStrFragmentEllipsis = "...     "
+
+	# ----
+
+	skipFrom = -1
+	skipTo = -1
+	if processorName:
+		if processorName == "shorten":
+			skipFrom = settings.compactBytesLinesLengthLimit
+			skipTo = nTotalLines - 4
+			if skipFrom >= skipTo:
+				skipFrom = -1
+				skipTo = -1
+		else:
+			raise Exception("No such postprocessor: " + repr(processorName))
+
+	# ----
+
+	if skipFrom < 0:
+		# direct loop, no addtional if statements
+		iFrom = 0
+		iTo = chunkSize
+		while iFrom < len(data):
+			chunk = data[iFrom:iTo]
+			yield formatStrFragment.format(iFrom), _x_byteChunkToHex(chunk).ljust(hexStrPadded), _x_byteChunkToASCII(chunk)
+			iFrom = iTo
+			iTo += chunkSize
+	else:
+		# loop with if statements
+		lineNo = 0
+		iFrom = 0
+		iTo = chunkSize
+		while iFrom < len(data):
+			chunk = data[iFrom:iTo]
+			if skipFrom <= lineNo <= skipTo:
+				if skipFrom == lineNo:
+					yield formatStrFragmentEllipsis, "...".ljust(hexStrPadded), "..."
+			else:
+				yield formatStrFragment.format(iFrom), _x_byteChunkToHex(chunk).ljust(hexStrPadded), _x_byteChunkToASCII(chunk)
+			iFrom = iTo
+			iTo += chunkSize
+			lineNo += 1
+#
+
+
 
 
 
@@ -110,6 +223,14 @@ class _Omitted:
 	pass
 #
 
+class RawValue:
+
+	def __init__(self, text:str) -> None:
+		self.text = text
+	#
+
+#
+
 _OMITTED = _Omitted()
 
 
@@ -118,7 +239,7 @@ _OMITTED = _Omitted()
 
 class DumpCtx(object):
 
-	_TYPE_MAP = {}
+	_TYPE_MAP = {}				# type -> function
 
 	def __init__(self, s:DumperSettings, outputLines:list, exitAppend:str, prefix:str):
 		self.__s = s
@@ -138,8 +259,12 @@ class DumpCtx(object):
 		self._dumpX(varName + " = ", value, processorName)
 	#
 
+	def dumpVarRaw(self, varName:str, value:RawValue) -> None:
+		self._dumpX(varName + " = ", value.text)
 	#
-	# ????
+
+	#
+	# This method is invoked if an object implements _dumpVarNames()
 	#
 	def dumpVars(self, caller, *args):
 		if len(args) == 0:
@@ -188,6 +313,14 @@ class DumpCtx(object):
 			self._dumpPrimitive(extraPrefix, None, processorName)
 			return
 
+		# is it a raw value?
+
+		if isinstance(value, RawValue):
+			if processorName is not None:
+				raise Exception("Raw values can not have processors.")
+			self._dumpRawValue(extraPrefix, value)
+			return
+
 		# is it one of our types?
 
 		t = type(value)
@@ -298,6 +431,29 @@ class DumpCtx(object):
 			self.outputLines.append(self.prefix + "]")
 	#
 
+	#
+	# Dump the specified byte array.
+	#
+	def _dumpBytes(self, extraPrefix:str, value:bytes, processorName:str = None):
+		e = (type(value).__name__ + ":") if self.__s.showComplexStructsWithType else ""
+
+		if len(value) <= self.__s.bytesLineSize:
+			self.outputLines.append(self.prefix + extraPrefix + e + repr(value))
+
+		else:
+			self.outputLines.append(self.prefix + extraPrefix + e + "<")
+
+			for sOfs, chunk, sAscii in _byteChunkerWithOfs(self.__s, value, processorName):
+				self.outputLines.append(self.prefix + "\t" + sOfs + "  " + chunk + "  " + sAscii)
+
+			if len(value) == 1:
+				self.outputLines.append(self.prefix + "\ttotal: 1 byte")
+			else:
+				self.outputLines.append(self.prefix + "\ttotal: " + str(len(value)) + " bytes")
+
+			self.outputLines.append(self.prefix + ">")
+	#
+
 	#
 	# Dump the specified tuple.
 	#
@@ -369,6 +525,10 @@ class DumpCtx(object):
 		self.outputLines.append(self.prefix + extraPrefix + self._primitiveValueToStr(value, processorName))
 	#
 
+	def _dumpRawValue(self, extraPrefix:str, value:RawValue):
+		self.outputLines.append(self.prefix + extraPrefix + value.text)
+	#
+
 	def _dumpOmitted(self, extraPrefix:str, value, processorName:str = None):
 		self.outputLines.append(self.prefix + extraPrefix + "...")
 	#
@@ -474,6 +634,7 @@ class DumpCtx(object):
 # Now let's register the types
 #
 if not DumpCtx._TYPE_MAP:
+	DumpCtx._TYPE_MAP[bytes] = DumpCtx._dumpBytes
 	DumpCtx._TYPE_MAP[set] = DumpCtx._dumpSet
 	DumpCtx._TYPE_MAP[frozenset] = DumpCtx._dumpFrozenSet
 	DumpCtx._TYPE_MAP[tuple] = DumpCtx._dumpTuple
@@ -518,6 +679,10 @@ class Dumper(object):
 			printFunc(line)
 	#
 
+	def toStr(self) -> str:
+		return "\n".join(self.__outputLines)
+	#
+
 #
 
 
@@ -528,7 +693,7 @@ class DumpMixin:
 
 	__slots__ = tuple()
 
-	def dump(self, prefix:str = None, printFunc = None):
+	def dump(self, prefix:str = None, printFunc = None) -> None:
 		dumper = Dumper()
 		with dumper.createContext(self, prefix) as dumper2:
 			if not dumper2._isDumpableObj(self):
@@ -537,6 +702,30 @@ class DumpMixin:
 		dumper.print(printFunc)
 	#
 
+	def dumpToStr(self, prefix:str = None) -> str:
+		dumper = Dumper()
+		with dumper.createContext(self, prefix) as dumper2:
+			if not dumper2._isDumpableObj(self):
+				raise Exception("Improper object encountered for prettyprinting: " + self.__class__.__name__ + " - Either implement _dump(ctx:DumpCtx) or _dumpVarNames()!")
+			dumper2._dumpObj("", self)
+		return dumper.toStr()
+	#
+
+	################################################################
+	# NOTE: Dumpable objects must implement one of these methods:
+	#
+	#	def _dump(self, ctx:DumpCtx):
+	#		ctx.dumpVar(...)
+	#	#
+	#
+	#	def _dumpVarNames(self) -> typing.List[str]:
+	#		return [
+	#			"....",
+	#		]
+	#	#
+	#
+	################################################################
+
 #
 
 
